# -*- coding: utf-8 -*-
"""
Created on Tue Jun 23 15:31:35 2020

@author: benjamin
Some functions used to read segmentation files
"""

from os import path
import numpy as np
import xml.etree.ElementTree as ET
import json

def removequotes(label):
    
    new_label = []
    for k1 in range(len(label)):
        lbltmp = label[k1]        
        new_label_tmp = []
        for k2 in range(len(lbltmp)):
            lbltmp2 = lbltmp[k2]
            if lbltmp2 == '"':
                lbltmp2 = '"#"'
            lbltmp2 = lbltmp2.replace('"', '')
            new_label_tmp.append(lbltmp2)
        new_label.append(new_label_tmp)
    return new_label

def func_readTextgrid(fileName):
    
    labels = []
    start = []
    stop = []
    
    POINT_BUFFER = 0.025

    if not path.exists(fileName):
        print('Error: %s not found\n', fileName)
        return labels, start, stop
    
    f = open(fileName, "r")
    C = np.loadtxt(f,
               delimiter='\n', dtype='str')
    f.close()
    
    tiers = -1
    proceed_int = False
    proceed_pnt = False
    xmin = 0
    xmax = 1

    for k in range(len(C)):
    
        A = C[k]
        if 'intervals: size ' in A.lower():
            tiers += 1        
            tier_len = int(A[A.index('=')+1:])
            labels.append([])
            start.append(np.zeros(tier_len))
            stop.append(np.zeros(tier_len))
            cnt = 0
            proceed_int = True
            
        if 'xmin' in A.lower():
            if proceed_int:
                start[tiers][cnt] = float(A[A.index('=')+1:])
            else:
                xmin = float(A[A.index('=')+1:])
                
        if 'xmax' in A.lower():
            if proceed_int:
                stop[tiers][cnt] = float(A[A.index('=')+1:])
            else:
                xmax = float(A[A.index('=')+1:])
                
        if 'text' in A.lower():
            if proceed_int:
                lab = A[A.index('=')+1:]
                while lab[0] == ' ':
                    lab = lab[1:]
                while lab[-1] != ('"'):
                    lab = lab[:-1]
                labels[tiers].append(lab)
                cnt += 1
                if cnt + 1 > tier_len:
                    proceed_int = False
                
        if 'points: size ' in A.lower():
            tiers += 1
            tier_len = int(A[A.index('=')+1:])
            labels.append([])
            start.append(np.zeros(tier_len))
            stop.append(np.zeros(tier_len))
            cnt = 0
            proceed_pnt = True
            
        if 'time' in A.lower():
            if proceed_pnt:
                start[tiers][cnt] = float(A[A.index('=')+1:]) - POINT_BUFFER
                stop[tiers][cnt] = float(A[A.index('=')+1:]) + POINT_BUFFER
        
        if 'mark' in A.lower():
            if proceed_pnt:
                lab = A[A.index('=')+1:]
                while lab[0] == ' ':
                    lab = lab[1:]
                while lab[-1] != ('"'):
                    lab = lab[:-1]
                labels[tiers].append(lab)
                cnt += 1
                if cnt + 1 > tier_len:
                    proceed_pnt = False
            
    return labels, start, stop

def getchildren(segm, word):
    idx = np.array([ x.tag.lower()==word.lower() for x in segm ]).astype(int)
    idx2 = np.argwhere(idx).reshape(-1).astype(int)
    return [segm[x] for x in idx2]

def read_xml_segmentation_files(fileName):
              
    start = []
    dur = []
    labels = []
    stop = []
    
    if not path.exists(fileName):
        print('Error: %s not found\n', fileName)
        return labels, start, stop
    
    root = ET.parse(fileName).getroot()
    listRoot = list(root)
    varTmp = getchildren(listRoot, 'SegmentList')
    segList = list(varTmp[0])
    structSegm = getchildren(segList, 'SpeechSegment')
    nSeg = len(structSegm)
        
    for kS in range(nSeg):
        varTmp = list(structSegm[kS])
    
        structWord = getchildren(varTmp, 'word') # varTmp(arrayfun(@(x)strcmpi(x.Name,'word'), varTmp));
        nWord = len(structWord) 
        for kW in range(nWord):
            currWord = structWord[kW]
            att = currWord.attrib
            readWord = currWord.text
            readWord = readWord.replace(' ', '')
            # readWord = readWord.replace('.', '')
            # readWord = readWord.replace(',', '')
            # readWord = readWord.replace('?', '')
            # readWord = readWord.replace(';', '')
            # readWord = readWord.replace('!', '')
            # readWord = readWord.replace('/', '')
            # readWord = readWord.replace(chr(8217), '')
            # readWord = readWord.replace(chr(39), '')

            if readWord != []:  
                idx_dur = att.get('dur')
                idx_start = att.get('stime')
                dur.append(float(idx_dur))
                start.append(float(idx_start))
                labels.append(readWord)
    
    stop = [x + y for (x,y) in zip(start,dur)]
    
    return labels, start, stop

def read_json_segmentation_files(fileName):
    
    labels = []
    instants = []    
   
    if not path.exists(fileName):
        print('Error: %s not found\n', fileName)
        return labels, instants
    
    with open(fileName, 'r') as myfile:
        jstr = myfile.read().replace('\n', '')
    idjson = json.loads(jstr)    
    tiers = idjson['tiers']
    nTiers = len(tiers)
    
    if nTiers == 1 and type(tiers) is not list:
        tiers = [tiers]        
    
    for k in range(nTiers):
        events = tiers[k]['events']
        if type(events) is dict:
            events = [events]
        nEvents = len(events)     
        lblTmp = []
        instTmp = np.zeros((nEvents, 2))
        for kE in range(nEvents):
            lblTmp.append(events[kE]['segment_name'])
            instTmp[kE,:] = [ events[kE]['segment_start'], 
                             events[kE]['segment_stop'] ]
            
        labels.append(lblTmp)
        instants.append(instTmp)
    
    return labels, instants

def read_seg_segmentation_files(fileName):
    
    labels = []
    instants = []
    
    if not path.exists(fileName):
        print('Error: %s not found\n', fileName)
        return labels, instants
    
    f = open(fileName, "r")
    C = np.loadtxt(f,
                   comments=None,
                   delimiter='\n', 
                   dtype='str')
    
    nPhon = len(C)
    lblPhon = []
    instPhon = []
    lblWord = []
    instWord = []
    lblSent = []
    instSent = []
    iterSent = 0
    
    for k in range(nPhon):
        currPhon = C[k]
        if '#@ sid=' in currPhon:
            idx1 = currPhon.find('-')
            idx2 = currPhon.find('-', idx1 + 1)
            sentStart = float(currPhon[idx1+1:idx2])
            sentEnd =  float(currPhon[idx2+1:])
            sentTmp = np.array([sentStart, sentEnd]).reshape(1,-1)
            
            if iterSent == 0:
                instSent = sentTmp
            else:
                lblSent.append(lblSentTmp)
                instSent = np.vstack((instSent, sentTmp))
    
            iterSent += 1
            lblSentTmp = []
        else:
            if '#@ word' in currPhon:
                idx1 = currPhon.find('=')
                idx2 = currPhon.find(' ')
                idx2 = currPhon.find(' ', idx2 + 1)
                Word = currPhon[idx1+1:idx2]
                lblWord.append(Word)
                if lblSentTmp == []:
                    lblSentTmp = Word
                else:
                    lblSentTmp = lblSentTmp + ' ' + Word
            
            if currPhon[0] != '#':
                idx1 = currPhon.find(' ')
                idx2 = currPhon.find(' ', idx1 + 1)
                idx3 = currPhon.find(' ', idx2 + 1)
                idx4 = currPhon.find(' ', idx3 + 1)
                idx5 = currPhon.find(' ', idx4 + 1)
                idx6 = currPhon.find(' ', idx5 + 1)

                fileId = currPhon[:idx1]
                sentId = currPhon[idx1+1:idx2]
                idx2Tmp = sentId.find(':')        
                label = sentId[idx2Tmp+2]
                sentPos = sentId[-1]
                start = float(currPhon[idx3+1:idx4]) / 100 + sentStart - 1e-2
                sumC = float(currPhon[idx4+1:idx5]) / 100
                sumC += float(currPhon[idx5+1:idx6]) / 100
                
                wEnd = start + sumC + float(currPhon[idx6+1:]) / 100

                lblPhon.append(label)
                instTmp = np.array([start, wEnd]).reshape(1,-1)
                if instPhon == []:
                    instPhon = instTmp
                else:
                    instPhon = np.vstack((instPhon, instTmp))
                if sentPos in ['1', '3']:
                    if instWord == []:
                        instWord = instTmp
                    else:
                        instWord = np.vstack((instWord, instTmp))
                if sentPos in ['2', '3']:
                    instWord[-1,1] = wEnd

    instants = [instPhon, instWord, instSent]
    labels = [lblPhon, lblWord, lblSent]     
   
    return labels, instants

def read_whc_segmentation_files(fileName):
    
    labels = []
    instants = []
    
    if not path.exists(fileName):
        print('Error: %s not found\n', fileName)
        return labels, instants
    
    f = open(fileName, "r")
    C = np.loadtxt(f,
               delimiter='\n', dtype='str')
    
    nWord = len(C)
    lblWord = []
    instWord = []
    for k in range(nWord):
        currWord = C[k]  
        if currWord[0] == '(':
            idx1 = currWord.find('-')
            idx2 = currWord.find('-', idx1 + 1)
            sentStart = float(currWord[idx1+1:idx2])
        else:
            idx1 = currWord.find(' ')
            idx2 = currWord.find(' ', idx1 + 1)
            idx3 = currWord.find(' ', idx2 + 1)
            label = currWord[:idx1]
            start = float(currWord[idx1+1:idx2]) / 100 + sentStart - 1 / 100
            wEnd = float(currWord[idx2+1:idx3]) / 100 + sentStart

            lblWord.append(label)
            instTmp = np.array([start, wEnd]).reshape(1, -1)
            if instWord == []:
                instWord = instTmp
            else:
                instWord = np.vstack((instWord, instTmp))

    instants = [instWord]
    labels = [lblWord]     
   
    return labels, instants



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat May 25 15:48:04 2019

@author: benjamin
"""

import numpy as np
import scipy

def symmetricifft(x, nfft=None):
    """ returns the IFFT by symmetric inverse Fourier transform """
    if len(x.shape) == 1:
        x = x.reshape(-1,1)
    xconj = np.conj(x[-2:0:-1,:])
    xconc = np.concatenate((x, xconj), 0)
    if nfft is None:
        nfft = len(xconc)
    y = np.real(np.fft.ifft(xconc, nfft, axis=0))

    return np.squeeze(y)

def window(nwin, win_type):
    
    if win_type == 'hann':
        return scipy.signal.hann(nwin)
    elif win_type == 'hanning':
        return scipy.signal.hanning(nwin)
    elif win_type == 'hamming':
        return scipy.signal.hamming(nwin)
    else:
        raise ValueError('Wrong type of window')

def lin_freq(f, F1, Fp1):
    """ returns linear relationship for piece-wise linear warping """
    f_1 = [x for x in f if x >= F1[0] and x<= F1[1]]
    return (Fp1[1] - Fp1[0]) / (F1[1] - F1[0]) * np.array(f_1 - F1[0]) + Fp1[0]

def freq2mel(freq):
    """ converts frequency in mel scale """
    return 2595 * np.log10(1 + freq / 700)

def mel2freq(mel):
    """ converts mels to frequency scale """
    return 700 * (10**(mel / 2595) - 1)

def mel_filter_bank(sr, fmin=0, fmax=None, nb_filter=24, nfft=1024):
    """ compute mel filter banks for MFCC computation """
    if fmax is None:
        fmax = int(sr / 2)
        
    fmin_mel = freq2mel(fmin)
    fmax_mel = freq2mel(fmax)

    mels = np.linspace(fmin_mel, fmax_mel, num=nb_filter+2)
    mel_freqs = mel2freq(mels)
    
    center_freq = np.floor((nfft + 1) / sr * mel_freqs).astype(int)

    filters = np.zeros((len(center_freq)-2, int(nfft/2+1)))
    
    for n in range(len(center_freq)-2):
        filters[n, center_freq[n] : center_freq[n + 1]] = np.linspace(0, 1, center_freq[n + 1] - center_freq[n])
        filters[n, center_freq[n + 1] : center_freq[n + 2]] = np.linspace(1, 0, center_freq[n + 2] - center_freq[n + 1])
    
    enorm = 2.0 / (mel_freqs[2:nb_filter+2] - mel_freqs[:nb_filter])
    filters *= enorm[:, np.newaxis]
    
    return filters, mels, mel_freqs

def dct(dct_filter_num, filter_len):
    """ computes DCT basis """
    basis = np.empty((dct_filter_num,filter_len))
    basis[0, :] = 1.0 / np.sqrt(filter_len)
    
    samples = np.arange(1, 2 * filter_len, 2) * np.pi / (2.0 * filter_len)

    for i in range(1, dct_filter_num):
        basis[i, :] = np.cos(i * samples) * np.sqrt(2.0 / filter_len)
        
    return basis

def delta(x):
    nb_feat, nb_frame = x.shape
    y = np.zeros_like(x)
    for k in range(nb_frame):
        if k == 0:
            y[:,k] = 0.5 * x[:,k+1]
        elif k == nb_frame - 1:
            y[:,k] = - 0.5 * x[:,k-1]
        else:
            y[:,k] = 0.5 * (x[:, k+1] - x[:, k-1])
            
    return y

def dfw(x, f_input, form_input, form_output):
    """ frequency warping of cepstral envelope such that peaks match target formants """
    in_form = form_input
    in_form = np.insert(in_form, 0, 0)
    in_form = np.append(in_form, f_input[-1])

    out_form = np.insert(form_output, 0, 0)
    out_form = np.append(out_form, f_input[-1])

    for k in range(5):
        fTmp = lin_freq(f_input, in_form[k:k+2], out_form[k:k+2])
        if k == 0:
            f_output = fTmp
        else:
            f_output = np.vstack((f_output.reshape(-1,1), fTmp.reshape(-1,1)))

    f_output = f_output[:len(f_input)]

    f1 = scipy.interpolate.interp1d(f_input, x, kind='quadratic', fill_value='extrapolate')
    return f1(f_output.squeeze())

def ceps2peaks(x, freq, fxx):
    
    x = abs(x)
    indx = scipy.signal.argrelextrema(x, np.greater)
    freq_tmp = fxx[indx]
    try:
        out = np.sort([freq_tmp[np.argmin(abs(freq_tmp - x))] for x in freq])
    except:
        out = freq
        
    return out

def time_ceps_transform(Sxx, transOper, nfft):
    
    L = int(nfft / 2 + 1)
    nord = transOper.shape[0]
    ceps_coeff = symmetricifft(20 * np.log10(np.abs(Sxx)), nfft)[:nord, :]
    ceps_coeff = scipy.stats.zscore(ceps_coeff)
    cepsT = transOper @ ceps_coeff
    cepsT2 = scipy.stats.zscore(cepsT)
    Sxx_trans = 10**(np.real(np.fft.fft(cepsT2, nfft, axis=0))[:L, :] / 20)
    Sxx_in = 10**(np.real(np.fft.fft(ceps_coeff, nfft, axis=0))[:L, :] / 20)
    
    return Sxx_in, Sxx_trans

def freq_ceps_transform(Sxx, transOper, freq_out, fxx):
    
    Sxx_in = scipy.stats.zscore(Sxx)
    Sxx_trans = transOper @ Sxx_in
    Sxx_trans = scipy.stats.zscore(Sxx_trans)    
        
    return Sxx_in, Sxx_trans


def ceps_transform(Sxx, transOper, nwin, novlp, fxx=None, freq_out=None, method='linear'):
    """ cepstral transformation """
    
    nfft = nwin
    if freq_out is None:
        Sxx_in, Sxx_trans = time_ceps_transform(Sxx, transOper, nfft)
    else:
        Sxx_in, Sxx_trans = freq_ceps_transform(Sxx, transOper, freq_out, fxx)

    return Sxx_in, Sxx_trans

def get_order_ceps(f0, sr):
    """ returns the optimal cepstral order based on pitch estimation """

    if np.nanmedian(f0) < 0:
        f0 = 125
    else:
        f0 = np.nanmedian(f0)

    if f0 <= 0:
        f0 = 125
    CepOrder = 2 * np.round(sr / f0)

    if np.floor(CepOrder / 2) == CepOrder / 2:
        CepOrder -= 1

    return int(CepOrder)

def cepstral_filtering(logx, nwin, order=None):
    """ cepstral liftering """
    xRealCep = np.real(np.fft.ifft(logx))
    wzp = np.zeros_like(xRealCep)

    CepOrder = int(order)
    midOrdp1 = int((CepOrder+1)/2-1)
    midOrdm1 = int((CepOrder-1)/2)

    winCeps = scipy.signal.hann(CepOrder)
    wzp = np.concatenate((winCeps[midOrdp1:CepOrder],
                          np.zeros(nwin - CepOrder),
                          winCeps[:midOrdm1]))

    liftCeps = wzp * xRealCep
    xEnv = np.real(np.fft.fft(liftCeps, nwin, axis=0))

    return xEnv

def true_envelope(x, sr, toldB=2, order=None):
    """ returns the true envelope of x """
    nwin = len(x)
    Ao = np.log(abs(np.fft.fft(x)))
    Vo = -np.inf * np.ones_like(Ao)
    while any(20*np.log10(np.exp(Ao)/np.exp(Vo)) > toldB):
        Ao = np.maximum(Ao, Vo)
        Vo = cepstral_filtering(Ao, nwin, order)

    return Vo

def esprit(x, sr = 1, K = 12):
    """ estimate the frequency of the damped sinusoids that model x """
    
    idx = np.argmax(x)
    x = x[idx:]
    x = x - np.mean(x)
    
    M = int(min(100, np.floor(len(x) / 2)))
    Nl = len(x) - M + 1
    Nt = int(Nl / M)
    
    R = np.zeros((M,M))
    for k in range(Nt):
        deb = int(k * M)
        fin = int(deb + 2 * M - 1)
        xtmp = x[deb:fin]
    
        H = scipy.linalg.hankel(xtmp[0:M], xtmp[M - 1:])
        R += H.dot(H.T)
        
    u, s, d = np.linalg.svd(R)
    nx, ny = u.shape
    
    Up = u[1:,:K]
    Um = u[:-1,:K]
    Phi = np.linalg.pinv(Um).dot(Up)
    z,w = np.linalg.eig(Phi)
    freq = np.angle(z) / 2 / np.pi * sr
    
    return np.sort(freq[freq > 50])

def z2freqz(p, z, sr=2*np.pi, nfft=None):
    
    if nfft is None:
        nfft = 2**nextpow2(sr)
    L = int(nfft / 2 + 1)
    b = np.poly(z)
    a = np.poly(p)
    return scipy.signal.freqz(b, a, L, fs=sr)
    
def fill_nan(x):

    idxAll = np.arange(x.shape[0])
    idxVal = np.where(np.isfinite(x))
    f = scipy.interpolate.interp1d(idxAll[idxVal], x[idxVal], 
                                   kind='cubic', fill_value='extrapolate')
    return np.where(np.isfinite(x), x, f(idxAll))

def fill_nan_and_zero(x):
    
    y = fill_nan(x)
    first_non_zero = np.argwhere(y>0)[0][0]
    last_non_zero = len(y) - np.argwhere(y[::-1]>0)[0][0] - 1
    y[:first_non_zero] = y[first_non_zero]
    y[last_non_zero:] = y[last_non_zero]
    
    return y

def nextpow2(x):
    return int(np.ceil(np.log2(np.abs(x))))
    
def mat2column(x):
    return np.reshape(x, (np.size(x),1),'F')
    
def buffer(x, nwin, nhop):

    nX = len(x)
    
    idx1 = nhop
    nFrame = int((nX - nwin) / nhop) 
    xOut = np.zeros((nwin, nFrame))
    xOut[:,0] = x[:nwin]#.reshape(-1,1)
    
    # while idx1 + nwin < nX:
    for k in range(1,nFrame):
        idx2 = idx1 + nwin
        xTmp = x[int(idx1):int(idx2)]
        xOut[:,k] = xTmp
        idx1 += nhop

    return xOut

def reject_outliers(x, thresh=3):
    """ reject outliers in x. Outliers are considered as data that are 
    at a distance more than thresh*SMAD where SMAD is the 
    scaled median absolute deviation of x"""
    
    K = 1.4826
    numF, nFrame = x.shape    
    median = np.matlib.repmat(np.median(x, axis=1).reshape(-1,1), 1, nFrame)
    diff = np.median(np.abs(x - median), axis=1)
    med_abs_deviation = np.matlib.repmat((K * diff).reshape(-1,1), 1, nFrame)
    idx = abs(x - median) < thresh * med_abs_deviation
    x_out = None
    idx_true = []

    for k in range(nFrame):
        yTmp = idx[:,k]
        if yTmp.all():
            if x_out is None:
                x_out = np.array(x[:,k]).reshape(-1,1)
            else:
                x_out = np.hstack((x_out, 
                                   np.array(x[:,k]).reshape(-1,1)))
            idx_true.append(k)

    return x_out, idx_true

def ProdMat3D(A,B):
    
    # Compute the product of 3D matrices along a single dimension
    l1, l2, l3 = A.shape
    C = np.zeros_like(A) + 1j * np.zeros_like(A)
    if l1 == 2 and l2 == 2:    
        C11 = np.zeros((1,1,l3)) + 1j * np.zeros((1,1,l3))
        C12 = np.zeros((1,1,l3)) + 1j * np.zeros((1,1,l3))
        C21 = np.zeros((1,1,l3)) + 1j * np.zeros((1,1,l3))
        C22 = np.zeros((1,1,l3)) + 1j * np.zeros((1,1,l3))
        a11 = A[0,0,:]
        a12 = A[0,1,:]
        a21 = A[1,0,:]
        a22 = A[1,1,:]
        b11 = B[0,0,:]
        b12 = B[0,1,:]
        b21 = B[1,0,:]
        b22 = B[1,1,:]
        C11[0,0,:] = a11 * b11 + a12 * b21
        C21[0,0,:] = a21 * b11 + a22 * b21
        C12[0,0,:] = a11 * b12 + a12 * b22
        C22[0,0,:] = a21 * b12 + a22 * b22
        C = np.concatenate((np.concatenate((C11,C12),1),np.concatenate((C21,C22),1)),0);
    else:
        for k in range(l3):    
            C[:,:,k] = np.dot(A[:,:,k],B[:,:,k])        
    return C
    
def ChainMatrix(af, lf, freq, param, meth='tmm',  Tf=None):
    lw = len(freq)
    if Tf is None:
        Tf = np.tile(np.eye(2).reshape((2,2,1)),(1, 1, lw))

    A = np.zeros((1,1,lw)) + 1j * np.zeros((1,1,lw))
    B = np.zeros((1,1,lw)) + 1j * np.zeros((1,1,lw)) 
    C = np.zeros((1,1,lw)) + 1j * np.zeros((1,1,lw))
    Zo = af / (param.rho * param.c)

    if meth.lower() != 'cmp':
        om = param.freq * 2 * np.pi    
        S = 2 * np.sqrt(af * np.pi)        
        L = param.rho / af
        Celem = Zo / param.c        
    
    for k in range(len(af)-1,-1,-1):        
        if meth.lower() == 'cmp':
            if param.loss:
                argh = param.sig * lf[k] / param.c
            else:
                argh = 1j * om / param.c * lf[k]
                param.gam = 1
            A[0,0,:] = np.cosh(argh)
            B[0,0,:] = -1 / Zo[k] * param.gam * np.sinh(argh)
            C[0,0,:] = -Zo[k] / param.gam * np.sinh(argh)
        else:
            R = S[k]*np.sqrt(param.rho * param.mu * om) / (2 * np.sqrt(2) * af[k]**2)
            G = (param.adiabatic - 1) * S[k] / (param.rho * param.c**2) * np.sqrt(param.heat_cond * om / (2 * param.specific_heat * param.rho))
            if param.loss:
                invZw = 1 / (param.wr / S[k]**2 + 1j*om*param.wm / S[k]**2 + 1 / (1j * om * S[k]**2 / param.wc)) * param.wallyield
                gam = np.sqrt((R + 1j * om * L[k]) * (G + 1j * om * Celem[k] + invZw))
            else:
                gam = 1j * om / param.c

            A[0,0,:] = np.cosh(gam * lf[k])
            B[0,0,:] = -np.sinh(gam * lf[k]) / Zo[k]
            C[0,0,:] = -Zo[k] * np.sinh(gam * lf[k])            
        Tn = np.concatenate((np.concatenate((A,B),1),np.concatenate((C,A),1)),0)
        Tf = ProdMat3D( Tf, Tn )
    
    A = Tf[0,0,:].reshape(-1)
    B = Tf[0,1,:].reshape(-1)
    C = Tf[1,0,:].reshape(-1)
    D = Tf[1,1,:].reshape(-1)
    
    return A, B, C, D, Tf

def Cosine_interp( xi, yi, xo ):
 # Cosine interpolation at points xo

    xa = xi[0]
    xb = xi[-1] 
    nTubes, nPt = yi.shape
    if min(nTubes, nPt) == 1:
        yi = yi.squeeze()
        ya = yi[0]
        yb = yi[-1]
    else:
        ya = yi[:,0] 
        yb = yi[:,-1]
    dy = yb-ya
    Xo = (xo-xa)/(xb-xa)*np.pi+np.pi
    
    if dy.size > 1:
        [Xm, Ym] = np.meshgrid(Xo, dy)
        [ XMdum, Ya ] = np.meshgrid(Xo, ya)
    else:
        Ya = ya
        Xm = Xo
        Ym = dy
    return Ya+1/2*(1+np.cos(Xm))*Ym
    
def findelements(a, b):
    # find indices of elements in a that contains elements in b
    nB = len(b)
    idx = []
    
    for k in range(nB):
        bTmp = b[k]
        idxTmp = [i for i,x in enumerate(a) if x==bTmp]
        if idxTmp != []: 
            if idx == []:
                idx = idxTmp
            else:
                idx = np.concatenate((idx, idxTmp)).astype(int)
                
    return idx

def lininterp1(xi, yi, xo, axis=1):
    
    if type(xi) is list:
        f1 = scipy.interpolate.interp1d(xi,yi,'linear',fill_value='extrapolate')
    else:
        f1 = scipy.interpolate.interp1d(xi,yi,'linear',fill_value='extrapolate', axis = axis)
   
    return f1(xo)

def modiffric(x, nwin=512, novlp=None, filt_coeff=[1,-0.9]):
    
    if novlp is None:
        novlp = np.ceil(nwin * 3 / 4)
    nhop = nwin - novlp
    xbuff = buffer(x, nwin, nhop)
    
    nFrame = xbuff.shape[1]
    y = np.zeros_like(x)
    w = scipy.signal.hann(nwin)
    
    for k in range(nFrame):
        deb = int((k - 1) * nhop)
        fin = int(deb + nwin - 1)
        if fin > len(y):
            y = np.concatenate((y.reshape(-1), np.zeros(nwin)))

        idx = [x for x in range(deb,fin+1)]
        xwin = xbuff[:,k] * w
        x_new = scipy.signal.lfilter([1],filt_coeff, xwin)
        x_new = x_new * (np.std(xwin) / np.std(x_new))
        y[idx] += x_new

    return y[:len(x)]/1.5

def speechmodif(x, sr, phoneme, modif_coeff, labels, instants):
    
    envelope = np.ones(x.shape)
    if phoneme in ['p', 'k', 't']:
        isBurst = True
        isFric = False
    if phoneme in ['f', 'v']:
        isBurst = False
        isFric = True
    if phoneme not in ['f', 'v', 'p', 'k', 't']:
        isBurst = False
        isFric = False
        
    phonemePlus = [phoneme, phoneme + '_']
    idx = findelements(labels, phonemePlus)

    if idx == []:
        # print('Warning : phoneme ' + phoneme + ' not found')
        return x
    
    for k in range(len(idx)):    
        
        idxtmp = idx[k]
        idx_init = int(instants[idxtmp,0] * sr)
        idx_final = int(instants[idxtmp,1] * sr)
        if k !=len(idx)-1:
            if idx[k+1] == idxtmp + 1:                
                k = k+1
                idx_final = int(instants[idxtmp+1,1] * sr)
        idx_p = [x for x in range(idx_init, idx_final+1)]

        tr_dur = 0.01
        tr_size = np.floor(tr_dur * sr)
        if np.mod(tr_size, 2):
            tr_size = tr_size - 1

        win = scipy.signal.hann(int(tr_size))
        winup = win[:int(tr_size / 2)]
        windown = win[int(tr_size / 2 - 1):]
        if isBurst:
#              try
#                 yTmp = modifburst(x[idx_p], sr, modif_coeff)
#                 x[idx_p] = yTmp
#                   print('Burst enhancement working')
#             except
#                 print('Burst enhancement did not work')
                envelope[idx_p] = modif_coeff * np.ones(len(idx_p))
                rise_start = int(idx_p[0] - tr_size / 4 - 1)
                if rise_start < 0:
                    rise_start = 0
                rise_end = rise_start + len(winup)
                # rise_end = int(idx_p[0] - tr_size / 4 + len(winup)-1)
                envelope[rise_start:rise_end] = 1 + (modif_coeff-1) * winup

                fall_start = int(idx_p[-1] - tr_size / 4 - 1)
                fall_end = int(idx_p[-1] - tr_size / 4 + len(windown) - 1)
                envelope[fall_start:fall_end] = 1 + (modif_coeff - 1) * windown
#             end
        else:
            envelope[idx_p] = modif_coeff * np.ones(len(idx_p))   
            rise_start = int(idx_p[0] - tr_size / 4 - 1)
            # rise_end = int(idx_p[0] - tr_size / 4 + len(winup)-1)
            if rise_start < 0:
                rise_start = 0
            rise_end = rise_start + len(winup)
            envelope[rise_start:rise_end] = 1 + (modif_coeff-1) * winup
            fall_start = int(idx_p[-1] - tr_size / 4 - 1)
            fall_end = int(idx_p[-1] - tr_size / 4 + len(windown) - 1)
            envelope[fall_start:fall_end] = 1 + (modif_coeff - 1) * windown
            if isFric:
                nwin = 512
                xTmp = x[rise_start:fall_end] 
                xTmp2 = np.zeros(x.shape)
                xTmp2[:] = x
                xTmp2[rise_start:fall_end] = 0
                xTmp = np.concatenate((np.zeros(int(nwin / 2)).reshape(-1), 
                            xTmp.reshape(-1),
                            np.zeros(int(nwin / 2)).reshape(-1)))    
                yTmp = modiffric(xTmp, nwin, nwin * 3 / 4, [ 1, -0.9 ])
                idxStart = int(rise_start - nwin / 2)
                if idxStart < 0:
                    idxStart = 0
                idxStop = int(fall_end + nwin / 2)
                x2add = xTmp2[idxStart:idxStop]
                x[idxStart:idxStop] = x2add + yTmp[:len(x2add)]
 
    return x * envelope
#########################################################################
############## The following functions are from N. Ruty #################
### "N. Ruty. Modèles d'interactions fluide/parois dans le conduit vocal. 
### Applications aux voix et aux pathologies. PhD thesis, 2007 ##########
#########################################################################

def forcecompute(Ug,lg,d_Ug_dt,Ps,Psupra,xvec,h1,h2,A,B,h0,xs,hs,const):
    # calcul des 4 forces at the points i=0,1,2,3
    # l: reaction forces resulting from the pressure distribution in the flow channel LEFT from the point i
    # r: reaction forces resulting from the pressure distribution in the flow channel RIGHT from the point i
    # formules Annexe C Vilain, 
    
    
    # modification par NR le 02/04/07
    ############# New modifications by B. Elie, 2016
    #
    # d�veloppement limit� de Fr_h1 et Fl_h2, lorsque le rapport A[1]*x2/B[1] tend vers 0 (<pres ,
    # pres fix� dans init_gen) , ligne 78-89, 116-128, 193-206
    #
    # Fonction pour le calcul de la contribution des forces de pression sur les
    # masses 1 et 2
    # 
    # Param�tres d'entr�e:
    # Ug: D�bit glottique
    # Lg: largeur de la glotte,
    # d_Ug_dt: derivee du d�bit glottique
    # P: pression sous-glottique
    # Psupra: pression supra glottique
    # xvec: vecteur contenant les abscisses des masses
    # h1,h2: ouvertures au niveau des masses 1 et 2
    # A,B: coeficients des �quations des plaques
    # h0: ouverture en entr�e/sortie de constriction
    # Visco,Instat: indicateurs de prise en compte ou non de la viscosit� et de l'inertie
    # xs: abscisse du point de s�paration de l'�coulement
    # hs: ouverture au niveau du point de s�paration
    # 
    # Param�tres de sortie:
    # 
    # Fl_h2: force de pression sur la masse 2 � gauche
    # Fr_h2: force de pression sur la masse 2 � droite
    # Fl_h1: force de pression sur la masse 1 � gauche
    # Fr_h1: force de pression sur la masse 1 � droite
    
    rho = const.rho
    co_mu = const.mu
    sep = const.sep
    # initialisation des forces
    Fl_h2 = 0
    Fr_h1 = 0
    
    # assignation des abscisses � partir du vecteur xvec
    (x0, x1, x2, x3) = xvec
    pres = 1e-5
    # quelques variables temporaires
    tmp1  = 0.5*rho*(Ug**2)/(lg**2) #variable temporaire associ�e au terme de Bernoulli
    tmp2  = -12*co_mu*Ug/lg #variable temporaire associ�e au terme de Poiseuille
    tmp3  = -rho*d_Ug_dt/lg #variable temporaire li�e au terme d'inertie
    
    Fl_h1 = 0.5*(x1-x0)*lg*Ps #force sur la masse 1 � gauche, temporaire, valable en cas de fermeture
    
    if (h1>0) and (h2>0): 
        ########################################################
        ################## Fl_h1 #####################
        i1 = 1
        bern = tmp1*(1/(h0**2)-2*(Xv1(x0,x1,A,B,i1)-x0*Wv1(x0,x1,A,B,i1))/(x1-x0)**2) #     Bernoulli 
        pois = tmp2/(2*A[0])*(1/(h0**2)-2*(Xv1(x0,x1,A,B,i1)-x0*Wv1(x0,x1,A,B,i1))/(x1-x0)**2)#     Poiseuille
        inst = tmp3/A[0]*(2*(Zv1(x0,x1,A,B,i1)-x0*Yv1(x0,x1,A,B,i1))/(x1-x0)**2-np.log(h0)) #     Inertia   
        Fl_h1 = 0.5*(x1-x0)*lg*(Ps+bern+pois+inst)
        
        ########################################################
        ################## Fr_h1 #####################
        ### two cases :::: h2 < h1*sep and h2 > h1*sep ###
        if (h2 < h1*sep):
            if (abs(A[1]*x2/B[1])<pres): # Parrallel duct
                i2 = 2
    #             bern  = tmp1*(1/(h0**2)-1/(h1**2))
                bern = tmp1*(1/(h0**2)+2*(Xv1(x1,x2,A,B,i2)-x2*Wv1(x1,x2,A,B,i2))/(x2-x1)**2) #  Bernoulli
    #             pois = tmp2*(1/(2*A[0])*(1/(h0**2)-1/(h1**2))+(x2-x1)/(h1**3))
                pois = tmp2*(1/(2*A[0])*(1/(h0**2)-1/(h1**2))-1/B[1]**3*(x1-1.5*A[1]/B[1]*x1**2+2*A[1]**2/B[1]**2*x1**3)\
                    -2/(x2-x1)**2/B[1]**3*((x2**3/3-0.375*A[1]/B[1]*x2**4+0.4*A[1]**2/B[1]**2*x2**5-\
                                           x2*(x2**2/2-0.5*A[1]/B[1]*x2**3+0.5*A[1]**2/B[1]**2*x2**4))\
                                           -(x1**3/3-0.375*A[1]/B[1]*x1**4+0.4*A[1]**2/B[1]**2*x1**5-\
                                           x2*(x1**2/2-0.5*A[1]/B[1]*x1**3+0.5*A[1]**2/B[1]**2*x1**4)))) # Poiseuille
    #             inst = tmp3*(np.log(h1ia/h0)/A[0]+(x2-x1)/h1)
                inst = tmp3*(np.log(h1/h0)/A[0]-1/B[1]*(x1-0.5*A[1]/B[1]*x1**2+1/3*A[1]**2/B[1]**2*x1**3)-\
                    2/(x2-x1)**2/B[1]*((1/3*x2**3-1/8*A[1]/B[1]*x2**4+1/15*A[1]**2/B[1]**2*x2**5-\
                                      x2*(0.5*x2**2-1/6*A[1]/B[1]*x2**3+1/12*A[1]**2/B[1]**2*x2**4))\
                                      -(1/3*x1**3-1/8*A[1]/B[1]*x1**4+1/15*A[1]**2/B[1]**2*x1**5-\
                                      x2*(0.5*x1**2-1/6*A[1]/B[1]*x1**3+1/12*A[1]**2/B[1]**2*x1**4)))) # Inertia
                  #proposition modif par NR 11/04/2007
    #             inst = tmp3*(np.log(h1/h0)/A[0]+(x2-x1)/3/h1)            
            else: #glotte formant un canal convergent ou divergent avec h2<hs
                i2 = 2 #num�ro de la plaque sur laquelle on calcule les forces
                bern = tmp1*(1/(h0**2)+2*(Xv1(x1,x2,A,B,i2)-x2*Wv1(x1,x2,A,B,i2))/(x2-x1)**2)# Bernoulli
                pois = tmp2*(1/(2*A[0])*(1/(h0**2)-1/(h1**2))+1/(2*A[1])*(1/(h1**2)+2*(Xv1(x1,x2,A,B,i2)-x2*Wv1(x1,x2,A,B,i2))/(x2-x1)**2))# Poiseuille
                inst = tmp3*(np.log(h1/h0)/A[0] + 1/A[1]*(2*(x2*Yv1(x1,x2,A,B,i2)-Zv1(x1,x2,A,B,i2))/(x2-x1)**2-np.log(h1)))# Inertia
            
            Fr_h1 = 0.5*(x2-x1)*lg*(Ps+bern+pois+inst)
        else:   #cas o� la glotte forme un canal divergent avec h2>hs
            if (abs(A[1]*x2/B[1])<pres):
                i2 = 2 #num�ro de la plaque sur laquelle on calcule les forces
                tmpG = (xs-x1)/(x2-x1)*(x2-(xs+x1)/2) #variable temporaire
                bern = tmp1*(1/(h0**2)-(x2*Wv1(x1,xs,A,B,i2)-Xv1(x1,xs,A,B,i2))/(tmpG*(x2-x1)))
    #             terme de poiseuille
                pois = tmp2*(1/(2*A[0])*(1/(h0**2)-1/(h1**2))-1/B[1]**3*(x1-1.5*A[1]/B[1]*x1**2+2*A[1]**2/B[1]**2*x1**3)\
                    -2/(x2-x1)/tmpG/B[1]**3*((xs**3/3-0.375*A[1]/B[1]*xs**4+0.4*A[1]**2/B[1]**2*xs**5-\
                                           x2*(xs**2/2-0.5*A[1]/B[1]*xs**3+0.5*A[1]**2/B[1]**2*xs**4))\
                                           -(x1**3/3-0.375*A[1]/B[1]*x1**4+0.4*A[1]**2/B[1]**2*x1**5-\
                                           x2*(x1**2/2-0.5*A[1]/B[1]*x1**3+0.5*A[1]**2/B[1]**2*x1**4)))) 
    
    #             terme d'inertie
                inst = tmp3*(np.log(h1/h0)/A[0]-1/B[1]*(x1-0.5*A[1]/B[1]*x1**2+1/3*A[1]**2/B[1]**2*x1**3)\
                    -2/(x2-x1)/tmpG/B[1]*((1/3*xs**3-1/8*A[1]/B[1]*xs**4+1/15*A[1]**2/B[1]**2*xs**5-\
                                      x2*(0.5*xs**2-1/6*A[1]/B[1]*xs**3+1/12*A[1]**2/B[1]**2*xs**4))\
                                      -(1/3*x1**3-1/8*A[1]/B[1]*x1**4+1/15*A[1]**2/B[1]**2*x1**5-\
                                      x2*(0.5*x1**2-1/6*A[1]/B[1]*x1**3+1/12*A[1]**2/B[1]**2*x1**4))))
                
            else:
                i2 = 2 #num�ro de la plaque sur laquelle on calcule les forces
                tmpG = (xs-x1)/(x2-x1)*(x2-(xs+x1)/2) #variable temporaire       
        #         terme de bernoulli
                bern = tmp1*(1/(h0**2)-(x2*Wv1(x1,xs,A,B,i2)-Xv1(x1,xs,A,B,i2))/(tmpG*(x2-x1)))
        #         terme de poiseuille
                pois = tmp2*(1/(2*A[0])*(1/(h0**2)-1/(h1**2))+1/(2*A[1])*(1/(h1**2))-(x2*Wv1(x1,xs,A,B,i2)-Xv1(x1,xs,A,B,i2))/(tmpG*(x2-x1)))
        #         terme d'inertie
                inst = tmp3*(np.log(h1/h0)/A[0]+1/A[1]*((x2*Yv1(x1,xs,A,B,i2)-Zv1(x1,xs,A,B,i2))/(tmpG*(x2-x1))-np.log(h1))) 
            
            Fr_h1 = lg*tmpG*(Ps+bern+pois+inst)+lg*Psupra*(x2-xs)/(x2-x1)*(x2-(x2+xs)/2)
        ########################################################
        ################## Calcul de Fl_h2 #####################
        ### two cases :::: h2 < h1*sep et le cas h2 > h1*sep ###
        
        if (h2 < h1*sep):
            if (abs(A[1]*x2/B[1])<pres): #glotte formant un conduit parall�le
                i2 = 2
    #         terme de bernoulli
    #             bern  = tmp1*(1/(h0**2)-1/(h1**2))
                bern = tmp1*(1/(h0**2)-2*(Xv1(x1,x2,A,B,i2)-x1*Wv1(x1,x2,A,B,i2))/(x2-x1)**2)
    #         terme de poiseuille
                pois = tmp2*(1/(2*A[0])*(1/(h0**2)-1/(h1**2))-1/B[1]**3*(x1-1.5*A[1]/B[1]*x1**2+2*A[1]**2/B[1]**2*x1**3)\
                    +2/(x2-x1)**2/B[1]**3*((x2**3/3-0.375*A[1]/B[1]*x2**4+0.4*A[1]**2/B[1]**2*x2**5-\
                                           x1*(x2**2/2-0.5*A[1]/B[1]*x2**3+0.5*A[1]**2/B[1]**2*x2**4))\
                                           -(x1**3/3-0.375*A[1]/B[1]*x1**4+0.4*A[1]**2/B[1]**2*x1**5-\
                                           x1*(x1**2/2-0.5*A[1]/B[1]*x1**3+0.5*A[1]**2/B[1]**2*x1**4)))) 
    
    #         terme d'inertie
                inst = tmp3*(np.log(h1/h0)/A[0]-1/B[1]*(x1-0.5*A[1]/B[1]*x1**2+1/3*A[1]**2/B[1]**2*x1**3)+\
                    2/(x2-x1)**2/B[1]*((1/3*x2**3-1/8*A[1]/B[1]*x2**4+1/15*A[1]**2/B[1]**2*x2**5-\
                                      x1*(0.5*x2**2-1/6*A[1]/B[1]*x2**3+1/12*A[1]**2/B[1]**2*x2**4))\
                                      -(1/3*x1**3-1/8*A[1]/B[1]*x1**4+1/15*A[1]**2/B[1]**2*x1**5-\
                                      x1*(0.5*x1**2-1/6*A[1]/B[1]*x1**3+1/12*A[1]**2/B[1]**2*x1**4))))
                
            else: #glotte formant un canal convergent ou divergent avec h2<hs
                i2 = 2 #num�ro de la plaque sur laquelle on fait le calcul\
    #         terme de bernoulli
                bern = tmp1*(1/(h0**2)-2*(Xv1(x1,x2,A,B,i2)-x1*Wv1(x1,x2,A,B,i2))/(x2-x1)**2)
    #         terme de poiseuille	        
                pois = tmp2*(1/(2*A[0])*(1/(h0**2)-1/(h1**2))+1/(2*A[1])*(1/(h1**2)-2*(Xv1(x1,x2,A,B,i2)-x1*Wv1(x1,x2,A,B,i2))/(x2-x1)**2))
    #         terme d'inertie	        
                inst = tmp3*(np.log(h1/h0)/A[0] +1/A[1]*(2*(Zv1(x1,x2,A,B,i2)-x1*Yv1(x1,x2,A,B,i2))/(x2-x1)**2-np.log(h1)))
            
            Fl_h2 = 0.5*(x2-x1)*lg*(Ps+bern+pois+inst)
            
        else: #cas o� la glotte forme un canal divergent avec h2>hs
            if (abs(A[1]*x2/B[1])<pres): #glotte formant un conduit parall�le
                i2 = 2
                tmpH = (xs-x1)/(x2-x1)*((xs+x1)/2-x1) #variable temporaire
            #         terme de bernoulli
                bern = tmp1*(1/(h0**2)-(Xv1(x1,xs,A,B,i2)-x1*Wv1(x1,xs,A,B,i2))/(tmpH*(x2-x1)))
    #         terme de poiseuille
                pois = tmp2*(1/(2*A[0])*(1/(h0**2)-1/(h1**2))-1/B[1]**3*(x1-1.5*A[1]/B[1]*x1**2+2*A[1]**2/B[1]**2*x1**3)\
                    +2/(x2-x1)/tmpG/B[1]**3*((xs**3/3-0.375*A[1]/B[1]*xs**4+0.4*A[1]**2/B[1]**2*xs**5-\
                                           x1*(xs**2/2-0.5*A[1]/B[1]*xs**3+0.5*A[1]**2/B[1]**2*xs**4))\
                                           -(x1**3/3-0.375*A[1]/B[1]*x1**4+0.4*A[1]**2/B[1]**2*x1**5-\
                                           x1*(x1**2/2-0.5*A[1]/B[1]*x1**3+0.5*A[1]**2/B[1]**2*x1**4)))) 
    
    #         terme d'inertie
                inst = tmp3*(np.log(h1/h0)/A[0]-1/B[1]*(x1-0.5*A[1]/B[1]*x1**2+1/3*A[1]**2/B[1]**2*x1**3)\
                    +2/(x2-x1)/tmpH/B[1]*((1/3*xs**3-1/8*A[1]/B[1]*xs**4+1/15*A[1]**2/B[1]**2*xs**5-\
                                      x1*(0.5*xs**2-1/6*A[1]/B[1]*xs**3+1/12*A[1]**2/B[1]**2*xs**4))\
                                      -(1/3*x1**3-1/8*A[1]/B[1]*x1**4+1/15*A[1]**2/B[1]**2*x1**5-\
                                      x1*(0.5*x1**2-1/6*A[1]/B[1]*x1**3+1/12*A[1]**2/B[1]**2*x1**4))))
                
            else:
                i2 = 2 #num�ro de la plaque sur laquelle on fait le calcul\
                tmpH = (xs-x1)/(x2-x1)*((xs+x1)/2-x1) #variable temporaire
        #         terme de bernoulli
                bern = tmp1*(1/(h0**2)-(Xv1(x1,xs,A,B,i2)-x1*Wv1(x1,xs,A,B,i2))/(tmpH*(x2-x1)))
        #         terme de poiseuille
                pois = tmp2*(1/(2*A[0])*(1/(h0**2)-1/(h1**2))+1/(2*A[1])*(1/(h1**2))-(Xv1(x1,xs,A,B,i2)-x1*Wv1(x1,xs,A,B,i2))/(tmpH*(x2-x1)))
                #proposition NR, 11/04/2007
        #         pois = tmp2*(1/(2*A[0])*(1/(h0**2)-1/(h1**2))+1/(2*A[1])*(1/(h1**2)-(Xv1(x1,xs,A,B,i2)-x1*Wv1(x1,xs,A,B,i2))/(tmpH*(x2-x1))))
        #         terme d'inertie	
                inst = tmp3*(np.log(h1/h0)/A[0] +1/A[1]*((Zv1(x1,xs,A,B,i2)-x1*Yv1(x1,xs,A,B,i2))/(tmpH*(x2-x1))-np.log(h1)))
          
            Fl_h2 = lg*tmpH*(Ps+bern+pois+inst)+ lg*Psupra*(x2-xs)/(x2-x1)*((x2+xs)/2-x1)
            
    elif (h1>0) and (h2<=0):   #ajout pas NR (04/04/2007), pour calculer les forces de pression
        Fr_h1 = 0.5*(x2-x1)*lg*Ps #en cas de fermeture au niveau d'une masse ou l'autre
        Fl_h2 = 0.5*(x2-x1)*lg*Ps #ici, fermeture au niveau de la masse 2
    elif (h1<=0) and (h2>0):
        Fr_h1 = 0.5*(x2-x1)*lg*Psupra #ici fermeture au niveau de la masse 1
        Fl_h2 = 0.5*(x2-x1)*lg*Psupra
     #modif par NR et CBV, sortie de cette ligne de la condition if de d�part
        ########################################################
        ################## Calcul de Fr_h2 #####################
    
    Fr_h2 = 0.5*lg*Psupra*(x3-x2) 
    Fvf = np.array([ [Fl_h1, Fl_h2 ], [Fr_h1, Fr_h2 ]])         
    return Fvf

def Wv1(xi,xip1,A,B,i):

    #****************************************************#
    # Calcul de la primitive de 1/h2  , entre xi et x(i+1), pour 
    # la plaque numero i
    # 
    # Parametres d'entree
    # 
    # xi,xip1: abscisse entre lesquelles on calcule l'integrale
    # A,B: coefficient de l'equation du la plaque,
    # i: numero de la plaque
    # 
    # Parametre de sortie
    # 
    # X: valeur de l'integrale
    #****************************************************#                                
    # Formules de CV   
    #****************************************************#
    
    pres = 1e-5
    
    hi = A[i-1]*xi + B[i-1]  # = hi 
    hip1 = A[i-1]*xip1 + B[i-1]    # = hi+1 
    # 
    # if (hi == hip1)
    #     W = (xip1-xi)/(hi**2) #cas oe la plaque est horizontale
    if (abs(A[i-1]*xip1/B[i-1])<pres): #cas de presque horizontalite, avec dvpt limite
          W2 = 1/B[i-1]**2*(xip1-A[i-1]*xip1**2/B[i-1]+A[i-1]**2*xip1**3/B[i-1]**2)
          W1 = 1/B[i-1]**2*(xi-A[i-1]*xi**2/B[i-1]+A[i-1]**2*xi**3/B[i-1]**2) 
          W = W2-W1
    else:
        W = 1/A[i-1]*(1/hi-1/hip1)
        
    return W

def Xv1(xi,xip1,A,B,i):

    #****************************************************#
    # Calcul de la primitive de x/h2 , entre xi et x(i+1), pour 
    # la plaque numero i
    # 
    # Parametres d'entree
    # 
    # xi,xip1: abscisse entre lesquelles on calcule l'integrale
    # A,B: coefficient de l'equation du la plaque,
    # i: numero de la plaque
    # 
    # Parametre de sortie
    # 
    # X: valeur de l'integrale
    #****************************************************#
    
    pres = 1e-5
    hi = A[i-1]*xi+B[i-1]  # = hi 
    hip1 = A[i-1]*xip1+B[i-1]    # = hi+1 
    
    # if (hi == hip1)
    #     X = (xip1-xi)*xi/(hi**2)  #cas oe la plaque est horizontale
    
    if (abs(A[i-1]*xip1/B[i-1])<pres): #cas de presque horizontalite, avec dvpt limite
          X2 = 1/B[i-1]**2*(xip1**2/2-2*A[i-1]*xip1**3/3/B[i-1]+3*A[i-1]**2*xip1**4/4/B[i-1]**2)
          X1 = 1/B[i-1]**2*(xi**2/2-2*A[i-1]*xi**3/3/B[i-1]+3*A[i-1]**2*xi**4/4/B[i-1]**2) 
          X = X2-X1
    else:
        X = 1/(A[i-1]**2)*np.log(hip1/hi)+1/A[i-1]*(xi-hi/A[i-1])*(1/hi-1/hip1)
        
    return X

def Yv1(xi,xip1,A,B,i):

    #****************************************************#
    # Calcul de la primitive de ln(h)  , entre xi et x(i+1), pour 
    # la plaque numero i
    # 
    # Parametres d'entree
    # 
    # xi,xip1: abscisse entre lesquelles on calcule l'integrale
    # A,B: coefficient de l'equation du la plaque,
    # i: numero de la plaque
    # 
    # Parametre de sortie
    # 
    # X: valeur de l'integrale                              
    #****************************************************#
    
    pres = 1e-5
    hi = A[i-1]*xi+B[i-1]  # = hi 
    hip1 = A[i-1]*xip1+B[i-1]    # = hi+1 
    # if (hi == hip1)
    #     Y = (xip1-xi)*np.log(hi)  #cas oe la plaque est horizontale
    if (abs(A[i-1]*xip1/B[i-1])<pres): #cas de presque horizontalite, avec dvpt limite
        Y2=(xip1*np.log(B[i-1])+A[i-1]/B[i-1]/2*xip1**2-A[i-1]**2/B[i-1]**2/6*xip1**3)
        Y1=(xi*np.log(B[i-1])+A[i-1]/B[i-1]/2*xi**2-A[i-1]**2/B[i-1]**2/6*xi**3)
        Y=Y1-Y2
    else:
        Y = 1/A[i-1]*(hip1*np.log(hip1)-hip1-hi*np.log(hi)+hi)
    
    return Y

def Zv1(xi,xip1,A,B,i):
    
    #****************************************************#
    # Calcul de la primitive de x*ln(h)    , entre xi et x(i+1), pour 
    # la plaque numero i
    # 
    # Parametres d'entree
    # 
    # xi,xip1: abscisse entre lesquelles on calcule l'integrale
    # A,B: coefficient de l'equation du la plaque,
    # i: numero de la plaque
    # 
    # Parametre de sortie
    # 
    # X: valeur de l'integrale                            
    #****************************************************#
    
    pres = 1e-5
    hi = A[i-1]*xi+B[i-1]  # = hi 
    hip1 = A[i-1]*xip1+B[i-1]    # = hi+1 
    # if (hi == hip1)
    #     Z = xi*(xip1-xi)*np.log(hi)  #cas oe la plaque est horizontale
    if (abs(A[i-1]*xip1/B[i-1])<pres): #cas de presque horizontalite, avec dvpt limite
        Z2=(xip1**2/2*np.log(B[i-1])+A[i-1]*xip1**3/B[i-1]/3-A[i-1]**2*xip1**4/8/B[i-1]**2)
        Z1=(xi**2/2*np.log(B[i-1])+A[i-1]*xi**3/B[i-1]/3-A[i-1]**2*xi**4/8/B[i-1]**2)
        Z=Z2-Z1
    else:
        tempo = 1/(A[i-1]**2)
        tempo = tempo*(((hip1**2)*np.log(hip1)-(hi**2)*np.log(hi))/2-((hip1**2)-(hi**2))/4)
        Z     = tempo+1/A[i-1]*(xi-hi/A[i-1])*(hip1*np.log(hip1)-hip1-hi*np.log(hi)+hi)
        
    return Z

def solve_pol_2(a,b,c):
    
    delta = b*b-4*a*c
    
    if delta>=0:  
        x1 = (-b+np.sqrt(delta))/(2*a)
        x2 = (-b-np.sqrt(delta))/(2*a)   
    else:
        print('no real solution for this equation')   
    return max(x1,x2)

def powtr( x, a, b, dr=1 ):
    # converts area function to midsagittal distance, or the contrary according
    # ot the power transformation
    
    
    if dr == 1: # midsagittal to area function
        x = 100*x # x should be in cm
        y = a*(x**b)
        y = y*1e-4 # goes back to m^2
    else: # area to midsagittal distance
        x = 1e4*x # x should be in cm^2
        y = np.exp((np.log(x)-np.log(a))/b)
        y = y*1e-2 # goes back to m

    return y


# -*- coding: utf-8 -*-
"""
Created on Tue Jun 23 09:55:08 2020

@author: benjamin
"""

import json
import numpy as np
import numpy.matlib as npmt
from numpy.linalg import multi_dot
import scipy.signal as ssg
from .utils import ChainMatrix, mat2column, esprit, symmetricifft
# from .signaltools import esprit, symmetricifft
from ._set import struct, timestruct
from .vtnetwork import VtNetwork
from .areafunction import AreaFunction
from .acoustics import AcousticsTime, AcousticsFreq
from .plottools import plot_tf


class VtWaveguide(VtNetwork):
    """ Class for analyzing and make operations on vocal tract cavities """
    #### Properties
    area_function = None # area function
    formants = None # formants
    child_wvg = None # name of the child waveguide object
    parent_wvg = None # name of the parent waveguide object
    twin_wvg = None # name of the twin waveguide object
    anabranch_wvg = None # name of the anabranch waveguide object
    child_point = None # point of connection where the child waveguide is connected
    parent_point = None # point of connection of the parent waveguide
    radiation = None # Boolean. True if the waveguide radiates (e.g. oral tract).
    transfer_function = None # transfer function of the waveguide
    transfer_function_constriction = None
    chain_matrix = None
    chain_matrix_derivative = None
    input_impedance = None
    freq = None # frequency vector of the transfer function
    actualVT = None
    velopharyngeal_port = 0
    velum_area = 0
    norder = 1
    time_acoustics = AcousticsTime()
    freq_acoustics = AcousticsFreq()

    # Constructor method
    def __init__(self, *args):
        nargin = len(args)
        for k in range(0, nargin, 2):
            key, value = args[k:k+2]
            setattr(self, key, value)
            if key == 'area_function':
                self.area_function.parent = self

    def computetransferfunction(self, param=None, meth='tmm', loc=-1):
        """ Compute the transfer function and the impulse response of the vocal tract
        defined by area function Avt and lvt. The resonance frequencies are
        estimated via ESPRIT. They are the pole frequencies of the impulse
        response of the vocal tract. The constant values are stocked in
        param.
        The transfer function is computed using the chain paradigm by Sondhi and
        Schroeter (A hybrid time-frequency domain articulatory speech
        synthesizer, IEEE TASSP, 1987)"""

        if param is None:
            param = timestruct()
        if not hasattr(param, 'rho'):
            param.rho = 1.204/1000
        if not hasattr(param, 'c'):
            param.c = 343.4
        if not hasattr(param, 'mu'):
            param.mu = 1.9831e-5
        if not hasattr(param, 'a'):
            param.a = 130 * np.pi
        if not hasattr(param, 'b'):
            param.b = (30 * np.pi)**2
        if not hasattr(param, 'c1'):
            param.c1 = 4
        if not hasattr(param, 'wo2'):
            param.wo2 = (406 * np.pi)**2
        if not hasattr(param, 'c1n'):
            param.c1n = 72
        if not hasattr(param, 'heat_cond'):
            param.heat_cond = 0.0034
        if not hasattr(param, 'specific_heat'):
            param.specific_heat = 160
        if not hasattr(param, 'adiabatic'):
            param.adiabatic = 1.4
        if not hasattr(param, 'wallyield'):
            param.wallyield = True
        if not hasattr(param, 'loss'):
            param.loss = True

        param.freq[param.freq <= 1e-11] = 1e-11
        freq = param.freq.reshape(-1)
        af = self.area_function.area # to make it from the lips to the glottis
        lf = self.area_function.length # to make it from the lips to the glottis
        if af is None or lf is None:
            print('Warning: the waveguide instance has no area function!')
            return None
        w = 2 * np.pi * freq # angular frequency
        lw = len(w)

        if meth.lower() == 'cmp':
            param.alp = np.sqrt(1j*w * param.c1)
            param.bet = 1j*w * param.wo2 / ((1j*w + param.a) * 1j*w + param.b) + param.alp
            param.gam = np.sqrt((param.alp + 1j*w) / (param.bet + 1j*w))
            param.sig = param.gam * (param.bet + 1j*w)

        af_size = af.shape
        if len(af_size) == 1:
            nframe = 1
            af = af.reshape(-1, 1)
            lf = lf.reshape(-1, 1)
        else:
            nframe = af_size[1]
        transFun = np.zeros((lw, nframe)) + 1j*np.zeros((lw, nframe))
        Hf_cstr = np.zeros((lw, nframe)) + 1j*np.zeros((lw, nframe))

        if isinstance(loc, int) and nframe > 1:
            loc = loc * np.ones(nframe)
        if isinstance(loc, int):
            loc = [loc]

        for kf in range(nframe):
            Avt = af[:, kf]
            lvt = lf[:, kf]
            Grad = param.rho * w**2 / 2. / np.pi / param.c_s
            Jrad = 8 * param.rho * w / 3. / np.pi**(1.5) * Avt[-1]**(-0.5)
            Zrad = Grad + 1j*Jrad

            A, _, C, _, _ = ChainMatrix(Avt, lvt, freq, param, meth)

            Hf = 1 / (-C * Zrad + A)
            transFun[:, kf] = Hf
            Hf_cstr[:, kf] = Hf

            if loc[kf] > 0:
                param.lg = 0.03
                param.Ag0 = 0.4e-4
                Zg = 12 * param.mu * param.lg**3 / param.Ag0**3 + \
                0.875 * param.rho / 2 / param.Ag0**2 + \
                1j*freq * 2 * np.pi * param.rho * param.lg / param.Ag0
                k_tmp = loc[kf]
                aup = Avt[k_tmp::-1]
                adown = Avt[k_tmp+1:]
                lup = lvt[k_tmp::-1]
                ldown = lvt[k_tmp+1:]
                Aup, Bup, Cup, Dup, _ = ChainMatrix(aup, lup, freq, param, meth)
                Adown, Bdown, Cdown, Ddown, _ = ChainMatrix(adown, ldown, freq, param, meth)
                Tf_front = 1 / (Cdown * Zrad + Ddown)
                Z_front = (Adown * Zrad + Bdown) / (Cdown * Zrad + Ddown)
                Z_back = (Aup * Zg + Bup) / (Cup * Zg + Dup)
                Hf_cstr[:, kf] = Tf_front * Z_back / (Z_front + Z_back)

        self.transfer_function = np.squeeze(transFun)
        self.transfer_function_constriction = np.squeeze(Hf_cstr)
        self.freq = freq
        return transFun, freq

    def computeformants(self, param=None):
        """ Computes the resonance frequencies of the VT_Waveguide object.
         Use the ESPRIT method on the impulse response to estimate the poles of
         the transfer function"""

        if param is None:
            param = struct()
        if self.transfer_function is None:
            self.computetransferfunction(self, param)
        tf_size = self.transfer_function.shape
        if len(tf_size) == 1:
            nfreq = tf_size[0]
            nframe = 1
            self.transfer_function = self.transfer_function.reshape(-1, 1)
        else:
            nfreq, nframe = tf_size

        K = 2 * (param.nform + 2)
        fmt = np.empty((param.nform, nframe))
        fmt[:] = np.nan
        for kf in range(nframe):
            Hf = self.transfer_function[:, kf]
            if np.isinf(Hf[0]) or np.isnan(Hf[0]):
                Hf[0] = Hf[1]
            df = self.freq[2] - self.freq[1]
            x = symmetricifft(Hf)
            x = len(x) * x
            himp = x[:int(np.ceil(0.5 * len(x)))]
            himp -= np.mean(himp)
            fk = esprit(himp, df * 2 * (nfreq - 1), K)
            fmt[:, kf] = fk[:param.nform]
        self.formants = fmt
        return fmt

    def computenoise(self, Const, namp, Udc, aglot):
        """ compute the noise component inside the waveguide"""

        k = Const.k_iter
        af = self.area_function.area
        nTube, nFrame = af.shape
        A2o = np.append(aglot, af[:, k].squeeze())

        Nsk = np.zeros(nTube + 1)
        if self.actualVT is None:
            self.actualVT = np.arange(nTube).astype(int)
        actVT = np.array(self.actualVT)
        actVT2 = np.append(actVT, actVT[-1] + 1)
        idxAct = actVT2.reshape(-1, 1)

        ac = min(af[actVT.astype(int), k])
        ic = np.argmin(af[actVT.astype(int), k]).squeeze()
        fny = Const.fs/2
        if namp != 0:
            W = self.time_acoustics.noise_matrix
            dk = 20
            if k - dk >= 0:
                idxStart = k-dk
            else:
                idxStart = 0
            if k+dk <= nFrame:
                idxEnd = k + dk
            else:
                idxEnd = nFrame

            bb = np.arange(idxStart, idxEnd+1)

            if ic >= 38:
                hl = Const.hld
            else:
                fpeak = min(0.3*np.sqrt(np.pi)*Udc/(ac*np.sqrt(ac)), fny*0.999)
                ordl = 9+2*(38-ic-1)
                hl = ssg.firwin(int(ordl), float(fpeak/fny))

            Wb = W[idxAct, bb]
            if ic >= 38:
                Wb = ssg.lfilter(Const.hld2, 1, Wb)
            else:
                Wb = ssg.lfilter(hl, 1, Wb)
            Wb = Wb / np.std(Wb)
            phik = namp * (self.reynolds_inst**2 - Const.Rec**2)
            WbTmp = phik * (Wb[idxAct, round(len(bb) / 2)].squeeze())

            Nsk[actVT2] = WbTmp / (A2o[idxAct].squeeze()**2.5) * Udc**3
            if isinstance(self, MainOralTract) and Const.dynterm:
                self.time_acoustics.elements.Udj[ic+1] = Const.beta * phik
            Nsk[0] = 0
            if len(Nsk) > len(actVT2):
                Nsk[len(actVT2):] *= 0
        else:
            Nsk = np.zeros(nTube+1)
        self.time_acoustics.elements.Ns = Nsk

    def exportaf2json(self, fileName):
        """export the area function in a .JSON file"""

        af = self.area_function.area
        lf = self.area_function.length

        if hasattr(self, 'velopharyngeal_port'):
            vpo = self.velopharyngeal_port
        else:
            vpo = 0
        if hasattr(self, 'velum_area'):
            vArea = self.velum_area
        else:
            vArea = np.zeros_like(af)

        if len(af.shape) == 1:
            nFrame = 0
            nTubes = len(af)
        else:
            nTubes, nFrame = af.shape

        idx = fileName.find('.json')
        if idx >= 0:
            fileName = fileName[:idx]

        if isinstance(vpo, (float, int)) and nFrame > 0:
            vpo = vpo * np.ones(nFrame)

        if isinstance(vArea, (float, int)) and nFrame > 0:
            vArea = vArea * np.ones((nTubes, nFrame))

        if isinstance(vArea, (float, int)) and nFrame > 0:
            vArea = npmt.repmat(vArea.reshape(-1, 1), 1, nFrame)

        if nFrame == 0:
            fTmp = fileName + '.json'
            nFrame = 1
            vpo = [vpo]
            vArea = vArea * np.ones((nTubes, 1))
            af = af.reshape(-1, 1)
            lf = lf.reshape(-1, 1)
        for kF in range(nFrame):
            if nFrame > 1:
                fTmp = fileName + '_' + '%.5d' %(kF+1) + '.json'
            else:
                fTmp = fileName + '.json'

            idjson = {}
            idjson['VelumOpeningInCm'] = np.sqrt(vpo[kF]) * 1e2
            idjson['frameId'] = fTmp
            idjson['tubes'] = []
            for kT in range(nTubes):
                idjson['tubes'].append({
                    'area' : af[kT, kF] * 1e4,
                    'velumArea' : vArea[kT, kF] * 1e4,
                    'x' : lf[kT, kF] * 1e2
                })
        with open(fTmp, 'w') as f:
            json.dump(idjson, f)

    def initacoustics(self):
        """ Initialisation of the derivative and integration terms
        Nx is the number of tubelets of the considered pipe"""

        self.time_acoustics = AcousticsTime()
        stmp = self.area_function.area.shape
        nTube = stmp[0]
        if len(stmp) == 1:
            nFrame = 1
        else:
            nFrame = stmp[1]

        P = np.zeros((nTube+1, nFrame))
        U = np.zeros((nTube+1, nFrame))
        W = np.random.randn(nTube+1, nFrame)

        if isinstance(self, GlottalChink):
            P = np.zeros((1, nFrame))
            U = np.zeros((1, nFrame))
            self.glottal_flow_vector = np.zeros(nFrame)
            self.glottal_flow_inst = 0

        if isinstance(self, MainOralTract):
            self.glottal_flow_vector = np.zeros(nFrame)
            self.glottal_flow_inst = 0
            self.reynolds_inst = 0
            self.reynolds_vector = np.zeros(nFrame)

        if isinstance(self, SubGlottalTract):
            self.connection_pressure_inst = 0
            self.connection_pressure_vector = np.zeros(nFrame)

        self.time_acoustics.pressure_time_matrix = P
        self.time_acoustics.flow_time_matrix = U
        self.time_acoustics.pressure_time = P[:, 0]
        self.time_acoustics.flow_time = U[:, 0]
        self.time_acoustics.parent = self

        if self.actualVT is None:
            self.actualVT = np.arange(nTube).astype(int)
        actVT = np.array(self.actualVT).astype(int)
        if nTube > max(actVT):
            W[max(actVT):, :] *= 0
        self.time_acoustics.noise_matrix = W

    def formant2af(self, formant_target, param=None):
        """Dynamic acoustic-to-articulatory inversion of oral vowels using the sensitivity
        matrix technique"""

        if param is None:
            param = struct()
        if not hasattr(param, 'length_var'):
            param.length_var = True
        if not hasattr(param, 'lips_var'):
            param.lips_var = True
        if not hasattr(param, 'area_weight'):
            param.area_weight = 0
        if not hasattr(param, 'length_weight'):
            param.length_weight = 0
        if not hasattr(param, 'threshold'):
            param.threshold = 1
        if not hasattr(param, 'laxNumIter'):
            param.maxNumIter = 1000
        if not hasattr(param, 'psi'):
            param.psi = 0.5
        if not hasattr(param, 'tf_method'):
            param.tf_method = 'cmp'
        if not hasattr(param, 'verbose'):
            param.verbose = True

        dl = param.length_var
        norm_lips = True - param.lips_var
        carea = param.area_weight
        cleng = param.length_weight
        cpot = param.potential_weight
        thresh = param.threshold
        Nbreak = param.maxNumIter
        # Acceleration term. psii > 1 => enhance speed,
        # but risk of unstability,
        # psii < 1 to make sure we are in the linearity domain...
        psii = param.psi
        err_form = np.zeros(int(Nbreak))

        # size of the acoustic vector
        shapeForm = formant_target.shape
        nFormant = shapeForm[0]
        if len(shapeForm) == 1:
            nFrame = 1
        else:
            nFrame = shapeForm[1]

        # get initial solutions
        a0 = self.area_function.area
        l0 = self.area_function.length
        shapeAf = a0.shape
        nTube = shapeAf[0]
        if len(shapeAf) == 1:
            nFrame = 1
            a0 = a0.reshape((nTube, nFrame), order='F')
            l0 = l0.reshape((nTube, nFrame), order='F')
        else:
            nFrame = shapeAf[1]

        # Vector must be column vectors
        a0vec = mat2column(a0)
        l0vec = mat2column(l0)
        ftargvec = mat2column(formant_target)

        # Initalization
        count_loop = 1
        ak = a0vec[:]
        lk = l0vec[:]
        lkp1 = lk[:]
        akp1mat = a0[:]
        lkp1mat = l0[:]

        if nFrame == 1:
            carea = 0
            cleng = 0
            # if only one frame, no kinetic energy constraints (static configuration)

        self.freq_acoustics.parent = self
        self.computetransferfunction(param, meth=param.tf_method)
        fk = self.computeformants(param)
        Sn_x = self.freq_acoustics.areasensitivityfunction(fk, param)
        if dl:
            Sn_x_l = self.freq_acoustics.lengthsensitivityfunction(fk, param)

            if nFrame > 1:
                Sn_x_l = Sn_x_l.reshape((nTube * nFrame, nFrame * nFormant),
                                        'F')
            else:
                Sn_x_l = np.squeeze(Sn_x_l)

        if nFrame > 1:
            Sn_x = Sn_x.reshape((nTube * nFrame, nFrame * nFormant),
                                'F')
        else:
            Sn_x = np.squeeze(Sn_x)

        fnest = mat2column(fk)

        currErr = np.mean(abs(fnest-ftargvec)/fnest)*100
        if param.verbose:
            print('Cost function is ' + '%.4f' %currErr)
        err_form = [currErr]


        while (err_form[-1] >= thresh and count_loop <= Nbreak):

            zn = (ftargvec-fnest)/fnest # Delta_f vector
            alph = np.real(multi_dot([zn.T, Sn_x.T, Sn_x, zn])/ \
                           multi_dot([multi_dot([Sn_x.T, Sn_x, zn]).T,
                                      Sn_x.T, Sn_x, zn])) * psii

            if dl:
                alphl = np.real(multi_dot([zn.T, Sn_x_l.T, Sn_x_l, zn])/ \
                           multi_dot([multi_dot([Sn_x_l.T, Sn_x_l, zn]).T,
                                      Sn_x_l.T, Sn_x_l, zn])) * psii

            dak = alph[0][0] * (Sn_x @ zn) # perturbation of area function, if unconstrained
            dakD = 0
            daikD = 0 # initalization of constraint terms

            if carea != 0: # if kinetic energy constraint on area function is enabled
                dpdamat = np.zeros_like(akp1mat)
                dpmat = np.diff(akp1mat, 1, 1)
                dpdamat[:, 0] = -2 * dpmat[:, 0]
                dpdamat[:, 1:-1] = 2 * dpmat[:, :-1] - 2 * dpmat[:, 1:]
                dpdamat[:, -1] = 2 * dpmat[:, -1]
                dpdamat2 = dpdamat*npmt.repmat(np.sum(abs(dpdamat)**2, 0), nTube, 1)
                dakD = mat2column(dpdamat2)

            if cpot != 0: # if potential energy constraint is enabled
                aRep = npmt.repmat(np.sum(abs((a0-akp1mat)/a0)**2, 0),
                                   nTube, 1)
                daikD_mat = 2 * (a0 - akp1mat) / a0 * aRep
                daikD = mat2column(daikD_mat)

            akp1 = abs(ak+np.diag(ak)*((1-carea)*dak+carea*dakD+cpot*daikD))
        #     Ind = find(akp1<=0.1e-4)
        #     akp1(Ind) = exp(log(ak(Ind)) + log(dak(Ind)+1))
            akp1mat = akp1.reshape((nTube, nFrame), order='F')
            if norm_lips:
                akp1mat[-1, :] = a0[-1, :]
                # if lip aperture is imposed

            akp1 = mat2column(akp1mat)
            ak = akp1[:]

            if dl: # Compute the new length function
                dlkD = 0
                dlikD = 0
                if cleng != 0:
                    dldamat = np.zeros(lkp1mat.shape)
                    dlmat = np.diff(np.cumsum(lkp1mat, 0), 1, 1)
                    dldamat[:, 0] = -2 * dlmat[:, 0]
                    dldamat[:, 1:-1] = 2 * dlmat[:, :-1] - 2 * dlmat[:, 1:]
                    dldamat[:, -1] = 2 * dlmat[:, -1]
                    dldamat2 = dldamat*npmt.repmat(np.sum(abs(dldamat)**2, 0), nTube, 1)
                    dlkD = mat2column(dldamat2) # np.reshape(dldamat2, (np.size(dldamat2, 1)))

                if cpot != 0:
                    lRep = npmt.repmat(np.sum(abs((l0-lkp1mat))**2, 0),
                                       nTube, 1)
                    dlikD_mat = 2 * (l0 - lkp1mat) * lRep
                    dlikD = mat2column(dlikD_mat)

                xi_vec = ((1 - cleng) * (Sn_x_l @ zn) + cleng * dlkD + cpot * dlikD) * alphl[0][0]
                argdiag = np.squeeze(1 / (1 + xi_vec))
                Xi_mat = np.diag(argdiag)
                lkp1 = abs(Xi_mat @ lk.reshape(-1, 1))
                lkp1[lkp1 * 40 > 0.2] = 0.2 / 40
                lk = lkp1[:]

            lkp1mat = lkp1.reshape((nTube, nFrame), order='F')

            # Inialization
            self.area_function = AreaFunction('area', akp1mat,
                                              'length', lkp1mat,
                                              'parent', self)
            self.computetransferfunction(param, param.tf_method)
            fk = self.computeformants(param)
            Sn_x = self.freq_acoustics.areasensitivityfunction(fk, param)
            if dl:
                Sn_x_l = self.freq_acoustics.lengthsensitivityfunction(fk, param)
                if nFrame > 1:
                    Sn_x_l = Sn_x_l.reshape((nTube * nFrame, nFrame * nFormant),
                                            'F')
                else:
                    Sn_x_l = np.squeeze(Sn_x_l)

            if nFrame > 1:
                Sn_x = Sn_x.reshape((nTube * nFrame, nFrame * nFormant),
                                    'F')
            else:
                Sn_x = np.squeeze(Sn_x)

            fnest = mat2column(fk)
            count_loop += 1

            currErr = np.mean(abs(fnest - ftargvec) / fnest) * 100
            err_form.append(currErr)
            if param.verbose:
                print('Cost function is ' + '%.4f' %currErr)

        akmat = akp1mat
        lkmat = lkp1mat
        self.area_function = AreaFunction('area', akmat.squeeze(),
                                          'length', lkmat.squeeze(),
                                          'parent', self)
        fnest = fnest.reshape((nFormant, nFrame),
                              order='F')
        Hnest = self.computetransferfunction(param)

        return akmat, lkmat, fnest, Hnest, err_form

    def plottransferfunction(self, sl=None, hf=None):
        """ plot the transfer function of the waveguide"""

        if self.transfer_function is None:
            print('Warning: No transfer function computed for this waveguide!'
                  ' Please compute the transfer function first')
            return None

        tf = self.transfer_function
        freq = self.freq


        return plot_tf(freq, tf, sl, hf)
    
    def check_instance(self, class_type):
        """ check instance """
        return eval('isinstance(self, ' + class_type + ')')

class GlottalChink(VtWaveguide):
    """Class for modeling a glottal chink"""
    glottal_flow_vector = None
    glottal_flow_inst = None
    chink_length = None

    def __init__(self, *args):
        nargin = len(args)
        for k in range(0, nargin, 2):
            key, value = args[k:k+2]
            setattr(self, key, value)
            if key == 'area_function':
                self.area_function.parent = self

        self.parent_point = [1]
        self.radiation = False


class SubGlottalTract(VtWaveguide):
    """Class for modeling the subglottal tract"""
    connection_pressure_inst = None
    connection_pressure_vector = None

    def __init__(self, *args):
        nargin = len(args)
        for k in range(0, nargin, 2):
            key, value = args[k:k+2]
            setattr(self, key, value)
            if key == 'area_function':
                self.area_function.parent = self

        self.parent_point = 1
        self.radiation = False

class MainOralTract(VtWaveguide):
    """Class for modeling the main oral tract"""

    #### Instantaneous acoustics
    subglottal_pressure = None
    supraglottal_pressure = None
    glottal_bernoulli = None
    glottal_resistance = None
    glottal_inductance = None
    glottal_flow_inst = None
    chink_flow_inst = 0
    reynolds_inst = None
    Qug = 0

    #### Output acoustic matrix
    subglottal_pressure_vector = None
    supraglottal_pressure_vector = None
    reynolds_vector = None
    glottal_flow_vector = None

    # Constructor method
    def __init__(self, *args):
        nargin = len(args)
        for k in range(0, nargin, 2):
            key, value = args[k:k+2]
            setattr(self, key, value)
            if key == 'area_function':
                self.area_function.parent = self

        self.parent_point = 1
        self.parent_wvg = None
        self.radiation = True

    def computereynolds(self, Udc=0, Const=None):
        """Compute the Reynolds number inside the constriction of the waveguide
        defined by a, l."""

        if Const is None:
            Const = timestruct()
        co_mu = Const.mu
        rho = Const.rho
        af = self.area_function.area
        shapeAf = af.shape
        nTube = shapeAf[0]
        if len(shapeAf) == 1:
            nFrame = 1
            af = af.reshape(-1, 1)
        else:
            nFrame = shapeAf[1]
        if hasattr(Const, 'k_iter'):
            a = af[:, Const.k_iter]
        else:
            Const.k_iter = np.array(0, nFrame)
            a = self.area_function

        # find the constriction geometry
        if self.actualVT is None:
            self.actualVT = np.arange(nTube).astype(int)
        actVT = self.actualVT

        Ack = min(a[actVT])
        Reyn = 2*rho/co_mu*Udc/np.sqrt(np.pi*Ack)
        self.reynolds_inst = Reyn
        return Reyn

    def updatesup(self, Const):
        """ Update the glottal upstream pressure"""

        Rj = self.time_acoustics.elements.Rj[0]
        Lj = self.time_acoustics.elements.Lj[0]
        Rcm = self.time_acoustics.elements.Rcm[0]

        Ug = self.glottal_flow_inst
        Uch = self.chink_flow_inst
        currP = self.time_acoustics.pressure_time[0]
        addP = (Rj + 2 / Const.T * Lj + Rcm)
        psup = currP + addP * (Ug + Uch) - self.Qug
        self.Qug = 4 / Const.T * Lj * (Ug + Uch) - self.Qug
        self.supraglottal_pressure = psup
        return psup



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct 26 13:56:52 2020

@author: benjamin
"""

import numpy as np
from .utils import nextpow2, symmetricifft
import scipy.signal as ssg

def ola_param(nwin, novlp=None, sr=None):
    
    if novlp is None:
            novlp = int(nwin * 3 / 4)
    novlp = int(novlp)
    if sr is None:    
        nfft = int(nwin)
    else:
        nfft = 2**nextpow2(sr)
        
    nhop = nwin - novlp
    return novlp, nfft, nhop

def init_ola(x, nwin, nhop):
    
    zeros2add = np.zeros((nhop, 1))
    x2 = np.vstack((zeros2add, x.reshape(-1, 1), zeros2add)).squeeze()
    x_out = np.zeros_like(x2)
       
    return x2, x_out

def check_index(x, k, nwin, nhop):
    
    w = ssg.hann(nwin)
    deb = int(k * nhop)
    fin = deb + nwin - 1
    isC = True
    if fin > len(x) - 1:
        fin = len(x) - 1
        isC = False
    idx = np.arange(deb, fin+1).astype(int)
    xwin = x[idx] * w[:len(idx)]
    
    return xwin, idx, isC

def spec_frame(x, k, nwin, nhop, nfft):
    
    L = int(nfft / 2 + 1)
    xwin, idx, isC = check_index(x, k, nwin, nhop)
    fft_xwin = np.fft.fft(xwin, nfft)[:L]
    
    return xwin, fft_xwin, isC, idx

def change_frame(fft_xwin, h_new, h_orig, nfft, idx, x_out, xwin, isChange=True):
    
    init_en = sum(abs(xwin)**2)
    if init_en != 0 or not np.isnan(init_en):   
        if isChange:
            fft_new = fft_xwin * abs(h_new) / abs(h_orig)
            x_new = symmetricifft(fft_new, nfft)
        else: 
            x_new = xwin
        
        x2add = x_new[:len(idx)]    
        new_en = sum(abs(x2add)**2)
        if new_en == 0 or np.isnan(new_en):            
            pass
        else:
            x_out[idx] = x_out[idx] + x2add * init_en / new_en       
    
    return x_out

def check_frame(k, isC, maxFrame):
    
    k += 1
    if k >= maxFrame:
        isC = False
        
    return k, isC


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct 13 11:25:44 2020

@author: benjamin
"""

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import scipy.spatial as ssp

# Customize matplotlib
matplotlib.rcParams.update({'text.usetex': True,
                            'text.latex.preamble': '\\usepackage{tipa}'})


def special_characters(x):
    y = x
    y = y.replace("x", "\\textipa{@}")
    y = y.replace("#", "\#")
    y = y.replace("_", "\_")
    y = y.replace("~", "\~")
    y = y.replace("Z", "\\textipa{Z}")
    y = y.replace("S", "\\textipa{S}")
    y = y.replace("O", "\\textipa{O}")
    y = y.replace("^", "\\textipa{2}")
    y = y.replace("I", "\\textipa{I}")
       
    return y

def plot_segmentation(ax, lbl, inst, limy):
    for k in range(len(lbl)):
        ax.plot(inst[k,1] * np.ones((2,1)), limy,
             linestyle='--', color=[ 0, 0.498, 0 ], linewidth=1)
        plt.text(np.mean(inst[k,:]), limy[1], special_characters(lbl[k]),
            color=[ 0, 0.498, 0 ], fontsize=20, fontweight='bold',
            horizontalalignment='center', verticalalignment='top')

def extents(f):
    delta = f[1] - f[0]
    return [f[0] - delta / 2, f[-1] + delta / 2]

def plot_tf(freq, tf, sl=None, hf=None):

    if hf is None:
        h_out = plt.figure()
    else:
        # h_out = plt.figure(hf)
        plt.figure(hf.number)
        h_out = hf

    if sl is not None or len(tf.shape) == 1:
        if sl is not None:
            tf = tf[:, sl]
        plt.semilogy(freq, abs(tf))
        plt.xlabel('Frequency (Hz)', fontsize=14, fontweight='bold')
        plt.ylabel('Transfer function', fontsize=14, fontweight='bold')

    plt.show()

    return h_out

def plot_spec(Pxx, tsp, fsp, dyn, tvec, signal, lbl, inst,
              pitch_plot, pitch_time_points, pitch_values,
              formant_plot, formants_time, formants_freq):

    pmax = np.max(Pxx)
    pmaxmclim = pmax-dyn

    fig_handle = plt.figure()
    gs = fig_handle.add_gridspec(3, 1)
    ax1 = fig_handle.add_subplot(gs[:2, 0])
    ax1.set_title('Spectrogram', fontsize=14, fontweight='bold')
    plt.imshow(Pxx, aspect='auto', interpolation='none',
               extent=extents(tsp)+extents(fsp), origin='lower',
               clim=(pmaxmclim,pmax), cmap="gray_r")
    plt.ylabel('Frequency (Hz)', fontsize=14, fontweight='bold')
    plt.xlabel('Time (s)', fontsize=14, fontweight='bold')
    plt.xlim([0, max(tvec)])
    if lbl is not None:
        plot_segmentation(ax1, lbl[0], inst[0], [0, max(fsp)])
    if pitch_plot:
        ax1.plot(pitch_time_points, pitch_values, color='b', linewidth=3)
    if formant_plot:
        ax1.plot(formants_time, formants_freq, linewidth=2)
    plt.show()

    ax2 = fig_handle.add_subplot(gs[-1, 0], sharex=ax1)
    ax2.set_title('Signal', fontsize=14, fontweight='bold')

    plt.plot(tvec, signal)
    plt.xlabel('Time (s)', fontsize=14, fontweight='bold')
    plt.xlim([0, max(tvec)])
    miny = min(signal) * 1.1
    maxy = max(signal) * 1.1
    if lbl is not None:
        plot_segmentation(ax2, lbl[0], inst[0], [ -1e3,1e3 ])
    plt.ylim([miny, maxy])
    return fig_handle

def freq_plot(freq, Pxx, dyn):

    max_Pxx = np.max(abs(Pxx))
    fig_handle = plt.figure()
    plt.plot(freq, 10*np.log10(np.abs(Pxx)))
    plt.ylabel('Spectrum (dB)', fontsize=14, fontweight='bold')
    plt.xlabel('Frequency (Hz)', fontsize=14, fontweight='bold')
    plt.ylim([10*np.log10(np.abs(max_Pxx))-dyn, 10*np.log10(np.abs(max_Pxx))])
    plt.xlim([0, max(freq)])

    return fig_handle

def vowel_space(fk, n1=1, n2=2, h=None, k_plot=None):

    n = [n1-1, n2-1]
    convHull = ssp.ConvexHull(fk[n,:].T)

    if h is None:
        h = plt.figure()
        ax1 = plt.subplot(111)
    elif k_plot is not None:
        ax1 = plt.subplot(4, 4, k_plot)
    elif k_plot is None:
        ax1 = plt.subplot(111)

    plt.figure(h.number)
    ax1.plot(fk[n1-1, :], fk[n2-1, :], '.')
    # ax1.plot(fk[n1-1, convHull.simplices[:, 0]],
    #          fk[n2-1, convHull.simplices[:,1]], 'r')
    for simplex in convHull.simplices:
        ax1.plot(fk[n1-1, simplex], fk[n2-1, simplex],
                 'r', linewidth=2)
    plt.ylabel('$F' + str(n2) +'$', fontsize=14, fontweight='bold')
    plt.xlabel('$F' + str(n1) +'$', fontsize=14, fontweight='bold')
    ax1.set_title('$F' + str(n1) +'-F' + str(n2) + '$ space',
                  fontsize=14, fontweight='bold')

    return h, ax1

def plot_area(xval, yval, hf=None):

    if hf is None:
        h_out = plt.figure()
    else:
        h_out = hf
        plt.figure(hf.number)
    plt.plot(xval, yval)
    plt.xlabel('Distance from glottis (cm)', fontsize=14, fontweight='bold')
    plt.ylabel('Area (cm\\textsuperscript{2})', fontsize=14, fontweight='bold')

    return h_out

def pitch_plot(tp, y_pitch, y_contour, spl_tp, spl_v):

    h = plt.figure()
    plt.plot(tp, y_pitch)
    plt.plot(tp, y_contour)

    plt.xlabel('Time (s)', fontsize=14, fontweight='bold')
    plt.ylabel('Pitch (Hz)', fontsize=14, fontweight='bold')

    if spl_tp is not None:
        plt.plot(spl_tp, spl_v)

    return h

def contour_plot(tvec, y, tp, yI):

    h = plt.figure()
    plt.plot(tvec, y)
    plt.plot(tp, yI)
    plt.xlabel('Time (s)', fontsize=14, fontweight='bold')
    plt.ylabel('Intensity (dB)', fontsize=14, fontweight='bold')

    return h

def addplot(ax, tvec, y, color='r', lst='solid', lnw=2):
    ax.plot(tvec, y, color=color, linestyle=lst, linewidth=lnw)

def subplot_param(subax, tvec, y, lbl, inst, limy,
                  title, ylabel, xlabel=None, limx=None, ax_share=None):

    if ax_share is None:
        ax = plt.subplot(subax)
    else:
        ax = plt.subplot(subax, sharex=ax_share)
    ax.plot(tvec, y, linewidth=2)
    plt.title(title, fontsize=14, fontweight='bold')
    plt.ylabel(ylabel, fontsize=14, fontweight='bold')
    if xlabel is not None:
        plt.xlabel(xlabel, fontsize=14, fontweight='bold')
    miny = min([min(y) * 1.1, 0])
    maxy = min([limy[1], max(y) * 1.1])

    plot_segmentation(ax, lbl, inst, [miny, maxy])
    plt.ylim([miny, maxy])
    if plt.xlim is not None:
        plt.xlim(limx)

    return ax

def formantspace(objForm, subax, h=None, clr='b', order=1):

    if h is None:
        h = openfig()
    else:
        plt.figure(h.number)
    k_iter = 0

    for k1 in range(4):
        h, axTmp = pdfplot(objForm.stats[k1], n1=(k1+1),
                          subax=k1*5+1, hf=h, clr=clr)
        if order == 1:
            for k2 in range(k1):
                if k1 != k2:
                    h, axTmp = vowel_space(objForm.frequency, n1=(k2+1), n2=(k1+1),
                                  h=h, k_plot=subax[k_iter])
                    # ax.append(axTmp)
                    k_iter += 1
        else:
            for k2 in range(k1+1, 4):
                if k1 != k2:
                    h, axTmp = vowel_space(objForm.frequency, n1=(k1+1), n2=(k2+1),
                                  h=h, k_plot=subax[k_iter])
                    # ax.append(axTmp)
                    k_iter += 1

    return h

def hull_plot(fk, n1=1, n2=2, h=None, k_plot=None, clr='b'):
    
    n = [n1-1, n2-1]
    convHull = ssp.ConvexHull(fk[n,:].T)

    if h is None:
        h = plt.figure()
        ax1 = plt.subplot(111)
    elif k_plot is not None:
        ax1 = plt.subplot(3, 3, k_plot)
    elif k_plot is None:
        ax1 = plt.subplot(111)

    plt.figure(h.number)
    for simplex in convHull.simplices:
        ax1.plot(fk[n1-1, simplex], fk[n2-1, simplex],
                 color=clr, linewidth=2)
    plt.ylabel('$F' + str(n2) +'$', fontsize=14, fontweight='bold')
    plt.xlabel('$F' + str(n1) +'$', fontsize=14, fontweight='bold')
    ax1.set_title('$F' + str(n1) +'-F' + str(n2) + '$ space',
                  fontsize=14, fontweight='bold')

    return convHull.area

def hull_compare(objForm_1, objForm_2):
    
    h = openfig()
    k_iter = 1
    a_1 = []
    a_2 = []
    for k1 in range(4):
        for k2 in range(k1):
            if k1 != k2:
                a_1.append(hull_plot(objForm_1.frequency, n1=(k2+1), n2=(k1+1),
                              h=h, k_plot=k_iter))
                a_2.append(hull_plot(objForm_2.frequency, n1=(k2+1), n2=(k1+1),
                              h=h, k_plot=k_iter, clr='r')) 
                k_iter += 1
            
    barWidth = 0.4
    axTmp = plt.subplot(3,1,3)
    r1 = range(6)
    r2 = [x + barWidth for x in r1]
    axTmp.bar(r1, a_1, color='b', width = barWidth)
    axTmp.bar(r2, a_2, color='r', width = barWidth)
    plt.xticks([r + barWidth / 2 for r in range(6)],
                      ['$F_1-F_2$', '$F_1-F_3$', '$F_1-F_4$', 
                       '$F_2-F_3$', '$F_2-F_4$', '$F_3-F_4$'])
    plt.ylabel('Hull area (Hz\\textsuperscript{2})')

def pdfplot(x, n1, subax=None, hf=None, clr='b'):

    if hf is None:
        h_out = plt.figure()
    else:
        h_out = hf
        plt.figure(hf.number)

    if subax is not None:
        ax = plt.subplot(4,4,subax)
    else:
        ax = plt.subplot(111)
    ax.plot(x.bins[:-1], x.pdf, linewidth=2, color=clr)
    plt.ylabel('PDF($F_' + str(n1) + '$)', fontsize=14, fontweight='bold')
    plt.xlabel('$F_' + str(n1) + '$ (Hz)', fontsize=14, fontweight='bold')

    return h_out, ax

def formant_stats(objForm, obj2compare=None, label=None):

    h = openfig()
    barWidth = 0.4
    ylbl = ['Mean (Hz)', 'Variance (Hz\\textsuperscript{2})', 
            'Skewness', 'Kurtosis']

    for k in range(4):
        ax = plt.subplot(5,4,k+1)
        if obj2compare is not None:
            ax.boxplot([objForm.stats[k].samples, obj2compare.stats[k].samples],
                       notch=True)
        else:
            ax.boxplot(objForm.stats[k].samples, notch=True)
        plt.ylabel('$F_' + str(k+1) + '$ (Hz)', fontsize=14, fontweight='bold')
        if label is not None:
            plt.gca().xaxis.set_ticklabels(label)
    for k1 in range(4):
        momTmp = []
        momTmp_2 = []
        for k2 in range(4):
            momTmp.append(objForm.stats[k2].moments[k1])
            if obj2compare is not None:
                momTmp_2.append(obj2compare.stats[k2].moments[k1])
        axTmp = plt.subplot(5,1,k1+2)
        r1 = range(4)
        axTmp.bar(r1, momTmp, color='b', width = barWidth)
        if obj2compare is not None:
            r2 = [x + barWidth for x in r1]
            axTmp.bar(r2, momTmp_2, color='r', width = barWidth)
        plt.xticks([r + barWidth / 2 for r in range(4)],
                      ['$F_1$', '$F_2$', '$F_3$', '$F_4$'])
        plt.ylabel(ylbl[k1])


    return h

def openfig():
    return plt.figure()